# SPDX-License-Identifier: CC0-1.0

# Copyright (C) 2022 Wojtek Kosior <koszko@koszko.org>
#
# Available under the terms of Creative Commons Zero v1.0 Universal.

from . import config
from .serve import start, start_wsgi



# SPDX-License-Identifier: AGPL-3.0-or-later

# Repository tests
#
# This file is part of Hydrilla
#
# Copyright (C) 2021, 2022 Wojtek Kosior
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# I, Wojtek Kosior, thereby promise not to sue for violation of this
# file's license. Although I request that you do not make use this code
# in a proprietary program, I am not going to enforce this in court.

# Enable using with Python 3.7.
from __future__ import annotations

import pytest
import sys
import shutil
import json
import functools as ft

from pathlib import Path
from hashlib import sha256
from tempfile import TemporaryDirectory
from typing import Callable, Optional

from flask.testing import FlaskClient
from markupsafe import escape
from werkzeug import Response

from hydrilla import util as hydrilla_util
from hydrilla.builder import build
from hydrilla.server import config, _version, serve

from .helpers import *

here        = Path(__file__).resolve().parent
config_path = here / 'config.json'
source_path = here / 'source-package-example'

expected_generated_by = {
    'name': 'hydrilla.server',
    'version': _version.version
}

SetupMod = Optional[Callable[['Setup'], None]]

source_files = (
    'index.json', 'hello.js', 'bye.js', 'message.js', 'README.txt',
    'README.txt.license', '.reuse/dep5', 'LICENSES/CC0-1.0.txt'
)

def run_reuse(command, **kwargs):
    """
    Instead of running a 'reuse' command, check if 'mock_reuse_missing' file
    exists under root directory. If yes, raise FileNotFoundError as if 'reuse'
    command was missing. If not, check if 'README.txt.license' file exists
    in the requested directory and return zero if it does.
    """
    expected = ['reuse', '--root', '<root>',
                'lint' if 'lint' in command else 'spdx']

    root_path = Path(process_command(command, expected)['root'])

    if (root_path / 'mock_reuse_missing').exists():
        raise FileNotFoundError('dummy')

    is_reuse_compliant = (root_path / 'README.txt.license').exists()

    return MockedCompletedProcess(command, 1 - is_reuse_compliant,
                                  stdout=f'dummy {expected[-1]} output',
                                  text_output=kwargs.get('text'))

@pytest.fixture
def mock_reuse(mock_subprocess_run):
    """
    Mock the REUSE command when executed through subprocess.run() from serve.py.
    """
    mock_subprocess_run(build, run_reuse)

class Setup:
    """
    Facilitate preparing test malcontent directory, Hydrilla config file and the
    actual Flask client. In a customizable way.
    """
    def __init__(self, modify_before_build: SetupMod=None,
                 modify_after_build: SetupMod=None) -> None:
        """Initialize Setup."""
        self._modify_before_build = modify_before_build
        self._modify_after_build = modify_after_build
        self._config = None
        self._client = None

    def _prepare(self) -> None:
        """Perform the build and call the callbacks as appropriate."""
        self.tmpdir = TemporaryDirectory()

        self.containing_dir = Path(self.tmpdir.name)
        self.malcontent_dir = self.containing_dir / 'sample_malcontent'
        self.index_json     = Path('index.json')

        self.source_dir = self.containing_dir / 'sample_source_package'
        for source_file in source_files:
            dst_path = self.source_dir / source_file
            dst_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(source_path / source_file, dst_path)

        self.config_path = self.containing_dir / 'config.json'
        shutil.copyfile(config_path, self.config_path)

        if self._modify_before_build:
            self._modify_before_build(self)

        build_obj = build.Build(self.source_dir, self.index_json)
        build_obj.write_package_files(self.malcontent_dir)

        if self._modify_after_build:
            self._modify_after_build(self)

    def config(self) -> dict:
        """Provide the contents of JSON config file used."""
        if self._config is None:
            self._prepare()
            self._config = config.load([self.config_path])

        return self._config

    def client(self) -> FlaskClient:
        """
        Provide app client that serves the objects from built sample package.
        """
        if self._client is None:
            app = serve.HydrillaApp(self.config(),
                                    flask_config={'TESTING': True})
            self._client = app.test_client()

        return self._client

def index_json_modification(modify_index_json):
    """Decorator for function that modifies index.json before build."""
    def handle_index_json(setup):
        """Modify index.json before build."""
        index_path = setup.source_dir / 'index.json'
        index_json, _ = hydrilla_util.load_instance_from_file(index_path)

        index_json = modify_index_json(index_json) or index_json

        index_json = f'''
        // SPDX-License-Identifier: CC0-1.0
        // Copyright (C) 2021, 2022 Wojtek Kosior
        {json.dumps(index_json)}
        '''

        index_path.write_text(index_json)

    return handle_index_json

@index_json_modification
def remove_all_uuids(index_json):
    """Modify sample packages to contain no (optional) UUIDs"""
    for definition in index_json['definitions']:
        del definition['uuid']

@index_json_modification
def bump_schema_v2(index_json) -> None:
    """Modify sample packages to use version 2 of Hydrilla JSON schemas."""
    for definition in index_json['definitions']:
        definition['min_haketilo_version'] = [1, 1]

        if definition['identifier'] == 'helloapple' and \
           definition['type'] == 'resource':
            definition['required_mappings'] = {'identifier': 'helloapple'}

default_setup = lambda: Setup()
uuidless_setup = lambda: Setup(modify_before_build=remove_all_uuids)
schema_v2_setup = lambda: Setup(modify_before_build=bump_schema_v2)

setup_makers = [default_setup, uuidless_setup, schema_v2_setup]

@pytest.mark.usefixtures('mock_reuse')
@pytest.mark.parametrize('setup_maker', setup_makers)
@pytest.mark.parametrize('item_type', ['resource', 'mapping'])
def test_get_newest(setup_maker, item_type: str) -> None:
    """
    Verify that
        GET '/{item_type}/{item_identifier}.json'
    returns proper definition that is also served at:
        GET '/{item_type}/{item_identifier}/{item_version}'
    """
    setup = setup_maker()
    response = setup.client().get(f'/{item_type}/helloapple.json')
    assert response.status_code == 200
    definition = json.loads(response.data.decode())
    assert definition['type']        == item_type
    assert definition['identifier']  == 'helloapple'

    response = setup.client().get(f'/{item_type}/helloapple/2021.11.10')
    assert response.status_code == 200
    assert definition == json.loads(response.data.decode())

    assert ('uuid' in definition) == (setup_maker is not uuidless_setup)

    hydrilla_util.validator_for(f'api_{item_type}_description-1.0.1.schema.json')\
                 .validate(definition)

@pytest.fixture
def setup(mock_reuse):
    """Prepare server test environment in the default way."""
    return default_setup()

def test_project_url(setup) -> None:
    """Fetch index.html and verify project URL from config is present there."""
    response = setup.client().get('/')
    assert b'html' in response.data
    project_url = setup.config()['hydrilla_project_url']
    assert escape(project_url).encode() in response.data

@pytest.mark.parametrize('item_type', ['resource', 'mapping'])
def test_get_nonexistent(setup, item_type: str) -> None:
    """
    Verify that attempts to GET a JSON definition of a nonexistent item or item
    version result in 404.
    """
    response = setup.client().get(f'/{item_type}/nonexistentapple.json')
    assert response.status_code == 404
    response = setup.client().get(f'/{item_type}/helloapple/1.2.3.999')
    assert response.status_code == 404

@pytest.mark.parametrize('item_type', ['resource', 'mapping'])
def test_file_refs(setup, item_type: str) -> None:
    """
    Verify that files referenced by definitions are accessible under their
    proper URLs and that their hashes match.
    """
    response = setup.client().get(f'/{item_type}/helloapple/2021.11.10')
    assert response.status_code == 200
    definition = json.loads(response.data.decode())

    for file_ref in [*definition.get('scripts', []),
                     *definition['source_copyright']]:
        hash_sum = file_ref["sha256"]
        response = setup.client().get(f'/file/sha256/{hash_sum}')

        assert response.status_code == 200
        assert sha256(response.data).digest().hex() == hash_sum

def test_empty_query(setup) -> None:
    """
    Verify that querying mappings for URL gives an empty list when there're no
    mathes.
    """
    response = setup.client().get(f'/query?url=https://nonexiste.nt/example')
    assert response.status_code == 200

    response_object = json.loads(response.data.decode())

    assert response_object == {
        '$schema': 'https://hydrilla.koszko.org/schemas/api_query_result-1.schema.json',
        'mappings': [],
        'generated_by': expected_generated_by
    }

    hydrilla_util.validator_for('api_query_result-1.0.1.schema.json')\
                 .validate(response_object)

def test_query(setup) -> None:
    """
    Verify that querying mappings for URL gives a list with reference(s) the the
    matching mapping(s).
    """
    response = setup.client().get(f'/query?url=https://hydrillabugs.koszko.org/')
    assert response.status_code == 200

    response_object = json.loads(response.data.decode())

    assert response_object == {
        '$schema': 'https://hydrilla.koszko.org/schemas/api_query_result-1.schema.json',
        'mappings': [{
            'identifier': 'helloapple',
            'long_name': 'Hello Apple',
            'version': [2021, 11, 10]
        }],
        'generated_by': expected_generated_by
    }

    hydrilla_util.validator_for('api_query_result-1.schema.json')\
                 .validate(response_object)

def test_source(setup) -> None:
    """Verify source descriptions are properly served."""
    response = setup.client().get(f'/source/hello.json')
    assert response.status_code == 200

    description = json.loads(response.data.decode())
    assert description['source_name'] == 'hello'

    assert sorted([d['identifier'] for d in description['definitions']]) == \
        ['hello-message', 'helloapple', 'helloapple']

    zipfile_hash = description['source_archives']['zip']['sha256']
    response = setup.client().get(f'/source/hello.zip')
    assert sha256(response.data).digest().hex() == zipfile_hash

    hydrilla_util.validator_for('api_source_description-1.schema.json')\
                 .validate(description)

def test_missing_source(setup) -> None:
    """Verify requests for nonexistent sources result in 404."""
    response = setup.client().get(f'/source/nonexistent.json')
    assert response.status_code == 404

    response = setup.client().get(f'/source/nonexistent.zip')
    assert response.status_code == 404

def test_normalize_version():
    assert hydrilla_util.normalize_version([4, 5, 3, 0, 0]) == [4, 5, 3]
    assert hydrilla_util.normalize_version([1, 0, 5, 0])    == [1, 0, 5]
    assert hydrilla_util.normalize_version([3, 3])          == [3, 3]



# SPDX-License-Identifier: AGPL-3.0-or-later

# Loading Hydrilla server configuration file.
#
# This file is part of Hydrilla
#
# Copyright (C) 2022 Wojtek Kosior
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# I, Wojtek Kosior, thereby promise not to sue for violation of this
# file's license. Although I request that you do not make use this code
# in a proprietary program, I am not going to enforce this in court.

# Enable using with Python 3.7.
from __future__ import annotations

import json

from pathlib import Path

import jsonschema

from .. import util

config_schema = {
    '$schema': 'http://json-schema.org/draft-07/schema#',
    'type': 'object',
    'properties': {
        'malcontent_dir': {
            'type': 'string'
        },
        'malcontent_dir': {
            'type': 'string'
        },
        'hydrilla_project_url': {
            'type': 'string'
        },
        'try_configs': {
            'type': 'array',
            'items': {
                'type': 'string'
            }
        },
        'use_configs': {
            'type': 'array',
            'items': {
                'type': 'string'
            }
        },
        'port': {
            'type': 'integer',
            'minimum': 0,
            'maximum': 65535
        },
        'werror': {
            'type': 'boolean'
        }
    }
}

here = Path(__file__).resolve().parent

def load(config_paths: list[Path]=[here / 'config.json'],
         can_fail: list[bool]=[]) -> dict:
    config = {}

    bools_missing = max(0, len(config_paths) - len(can_fail))
    config_paths = [*config_paths]
    can_fail = [*can_fail[:len(config_paths)], *([False] * bools_missing)]

    while config_paths:
        path = config_paths.pop()
        fail_ok = can_fail.pop()

        try:
            json_text = path.read_text()
        except Exception as e:
            if fail_ok:
                continue
            raise e from None

        new_config = json.loads(util.strip_json_comments(json_text))
        jsonschema.validate(new_config, config_schema)

        config.update(new_config)

        if 'malcontent_dir' in config:
            malcontent_dir = Path(config['malcontent_dir'])
            if not malcontent_dir.is_absolute():
                malcontent_dir = path.parent / malcontent_dir

        config['malcontent_dir'] = str(malcontent_dir.resolve())

        for key, failure_ok in [('try_configs', True), ('use_configs', False)]:
            paths = new_config.get(key, [])
            paths.reverse()
            config_paths.extend(paths)
            can_fail.extend([failure_ok] * len(paths))

    for key in ('try_configs', 'use_configs'):
        if key in config:
            config.pop(key)

    return config



# SPDX-License-Identifier: CC0-1.0

# Copyright (C) 2022 Wojtek Kosior <koszko@koszko.org>
#
# Available under the terms of Creative Commons Zero v1.0 Universal.

import sys
from pathlib import Path

import pytest
import pkgutil
import importlib
import functools
from tempfile import TemporaryDirectory
from typing import Iterable

here = Path(__file__).resolve().parent
sys.path.insert(0, str(here / 'src'))

@pytest.fixture(autouse=True)
def no_requests(monkeypatch):
    """Remove requests.sessions.Session.request for all tests."""
    if importlib.util.find_spec("requests") is not None:
        monkeypatch.delattr('requests.sessions.Session.request')

def _mock_subprocess_run(monkeypatch, where, mocked_run):
    """Temporarily replace subprocess.run() with the given function."""
    class MockedSubprocess:
        """Minimal mocked version of the subprocess module."""
        run = mocked_run

    monkeypatch.setattr(where, 'subprocess', MockedSubprocess)

@pytest.fixture
def mock_subprocess_run(monkeypatch, request):
    """
    Facilitate temporarily replacing subprocess.run() with a different function.

    If the 'subprocess_run' pytest marker has been used, perform the replacement
    for the module-function pair supplied through it.

    Return a function that can be called to perform the same replacement in
    another fixture or from inside a test function.
    """
    mocker = functools.partial(_mock_subprocess_run, monkeypatch)

    marker = request.node.get_closest_marker('subprocess_run')
    if marker:
        where, mocked_run = marker.args
        mocker(where, mocked_run)

    return mocker

@pytest.fixture(autouse=True)
def no_gettext(monkeypatch, request):
    """
    Make gettext return all strings untranslated unless we request otherwise.
    """
    if request.node.get_closest_marker('enable_gettext'):
        return

    import hydrilla
    modules_to_process = [hydrilla]

    def add_child_modules(parent):
        """
        Recursuvely collect all modules descending from 'parent' into an array.
        """
        try:
            load_paths = parent.__path__
        except AttributeError:
            return

        for module_info in pkgutil.iter_modules(load_paths):
            if module_info.name != '__main__':
                __import__(f'{parent.__name__}.{module_info.name}')
                modules_to_process.append(getattr(parent, module_info.name))
                add_child_modules(getattr(parent, module_info.name))

    add_child_modules(hydrilla)

    for module in modules_to_process:
        if hasattr(module, '_'):
            monkeypatch.setattr(module, '_', lambda message: message)

@pytest.fixture
def tmpdir() -> Iterable[Path]:
    """
    Provide test case with a temporary directory that will be automatically
    deleted after the test.
    """
    with TemporaryDirectory() as tmpdir:
        yield Path(tmpdir)



# SPDX-License-Identifier: AGPL-3.0-or-later

# Main repository logic.
#
# This file is part of Hydrilla
#
# Copyright (C) 2021, 2022 Wojtek Kosior
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# I, Wojtek Kosior, thereby promise not to sue for violation of this
# file's license. Although I request that you do not make use this code
# in a proprietary program, I am not going to enforce this in court.

# Enable using with Python 3.7.
from __future__ import annotations

import re
import os
import pathlib
import json
import logging

from pathlib import Path
from hashlib import sha256
from abc import ABC, abstractmethod
from typing import Optional, Union, Iterable

import click
import flask

from werkzeug import Response

from .. import util
from . import config
from . import _version

here = Path(__file__).resolve().parent

generated_by = {
    'name': 'hydrilla.server',
    'version': _version.version
}

class ItemInfo(ABC):
    """Shortened data of a resource/mapping."""
    def __init__(self, item_obj: dict, major_schema_version: int):
        """Initialize ItemInfo using item definition read from JSON."""
        self.version    = util.normalize_version(item_obj['version'])
        self.identifier = item_obj['identifier']
        self.uuid       = item_obj.get('uuid')
        self.long_name  = item_obj['long_name']

        self.required_mappings = []
        if major_schema_version >= 2:
            self.required_mappings = [map_ref['identifier'] for map_ref in
                                      item_obj.get('required_mappings', [])]

    def path(self) -> str:
        """
        Get a relative path to this item's JSON definition with respect to
        directory containing items of this type.
        """
        return f'{self.identifier}/{util.version_string(self.version)}'

class ResourceInfo(ItemInfo):
    """Shortened data of a resource."""
    def __init__(self, resource_obj: dict, major_schema_version: int):
        """Initialize ResourceInfo using resource definition read from JSON."""
        super().__init__(resource_obj, major_schema_version)

        dependencies = resource_obj.get('dependencies', [])
        self.dependencies = [res_ref['identifier'] for res_ref in dependencies]

class MappingInfo(ItemInfo):
    """Shortened data of a mapping."""
    def __init__(self, mapping_obj: dict, major_schema_version: int):
        """Initialize MappingInfo using mapping definition read from JSON."""
        super().__init__(mapping_obj, major_schema_version)

        self.payloads = {}
        for pattern, res_ref in mapping_obj.get('payloads', {}).items():
            self.payloads[pattern] = res_ref['identifier']

    def as_query_result(self) -> str:
        """
        Produce a json.dump()-able object describing this mapping as one of a
        collection of query results.
        """
        return {
            'version':    self.version,
            'identifier': self.identifier,
            'long_name':  self.long_name
        }

class VersionedItemInfo:
    """Stores data of multiple versions of given resource/mapping."""
    def __init__(self):
        self.uuid = None
        self.identifier = None
        self.by_version = {}
        self.known_versions = []

    def register(self, item_info: ItemInfo) -> None:
        """
        Make item info queryable by version. Perform sanity checks for uuid.
        """
        if self.identifier is None:
            self.identifier = item_info.identifier

        if self.uuid is None:
            self.uuid = item_info.uuid

        if self.uuid is not None and self.uuid != item_info.uuid:
            raise ValueError(f_('uuid_mismatch_{identifier}')
                             .format(identifier=self.identifier))

        ver = item_info.version
        ver_str = util.version_string(ver)

        if ver_str in self.by_version:
            raise ValueError(f_('version_clash_{identifier}_{version}')
                             .format(identifier=self.identifier,
                                     version=ver_str))

        self.by_version[ver_str] = item_info
        self.known_versions.append(ver)

    def get_by_ver(self, ver: Optional[list[int]]=None) -> Optional[ItemInfo]:
        """
        Find and return info of the newest version of item.

        If ver is specified, instead find and return info of that version of the
        item (or None if absent).
        """
        ver = util.version_string(ver or self.known_versions[-1])

        return self.by_version.get(ver)

    def get_all(self) -> list[ItemInfo]:
        """
        Return a list of item info for all its versions, from oldest ot newest.
        """
        return [self.by_version[util.version_string(ver)]
                for ver in self.known_versions]

class PatternTreeNode:
    """
    "Pattern Tree" is how we refer to the data structure used for querying
    Haketilo patterns. Those look like 'https://*.example.com/ab/***'. The goal
    is to make it possible for given URL to quickly retrieve all known patterns
    that match it.
    """
    def __init__(self):
        self.wildcard_matches = [None, None, None]
        self.literal_match    = None
        self.children         = {}

    def search(self, segments):
        """
        Yields all matches of this segments sequence against the tree that
        starts at this node. Results are produces in order from greatest to
        lowest pattern specificity.
        """
        nodes = [self]

        for segment in segments:
            next_node = nodes[-1].children.get(segment)
            if next_node is None:
                break

            nodes.append(next_node)

        nsegments = len(segments)
        cond_literal = lambda: len(nodes)     == nsegments
        cond_wildcard = [
            lambda: len(nodes) + 1 == nsegments and segments[-1] != '*',
            lambda: len(nodes) + 1 <  nsegments,
            lambda: len(nodes) + 1 != nsegments or  segments[-1] != '***'
        ]

        while nodes:
            node = nodes.pop()

            for item, condition in [(node.literal_match, cond_literal),
                                    *zip(node.wildcard_matches, cond_wildcard)]:
                if item is not None and condition():
                    yield item

    def add(self, segments, item_instantiator):
        """
        Make item queryable through (this branch of) the Pattern Tree. If there
        was not yet any item associated with the tree path designated by
        segments, create a new one using item_instantiator() function. Return
        all items matching this path (both the ones that existed and the ones
        just created).
        """
        node = self
        segment = None

        for segment in segments:
            wildcards = node.wildcard_matches

            child = node.children.get(segment) or PatternTreeNode()
            node.children[segment] = child
            node = child

        if node.literal_match is None:
            node.literal_match = item_instantiator()

        if segment not in ('*', '**', '***'):
            return [node.literal_match]

        if wildcards[len(segment) - 1] is None:
            wildcards[len(segment) - 1] = item_instantiator()

        return [node.literal_match, wildcards[len(segment) - 1]]

proto_regex  = re.compile(r'^(?P<proto>\w+)://(?P<rest>.*)$')
user_re      = r'[^/?#@]+@' # r'(?P<user>[^/?#@]+)@' # discarded for now
query_re     = r'\??[^#]*'  # r'\??(?P<query>[^#]*)' # discarded for now
domain_re    = r'(?P<domain>[^/?#]+)'
path_re      = r'(?P<path>[^?#]*)'
http_regex   = re.compile(f'{domain_re}{path_re}{query_re}.*')
ftp_regex    = re.compile(f'(?:{user_re})?{domain_re}{path_re}.*')

class UrlError(ValueError):
    """Used to report a URL or URL pattern that is invalid or unsupported."""
    pass

class DeconstructedUrl:
    """Represents a deconstructed URL or URL pattern"""
    def __init__(self, url):
        self.url = url

        match = proto_regex.match(url)
        if not match:
            raise UrlError(f_('invalid_URL_{}').format(url))

        self.proto = match.group('proto')
        if self.proto not in ('http', 'https', 'ftp'):
            raise UrlError(f_('disallowed_protocol_{}').format(proto))

        if self.proto == 'ftp':
            match = ftp_regex.match(match.group('rest'))
        elif self.proto in ('http', 'https'):
            match = http_regex.match(match.group('rest'))

        if not match:
            raise UrlError(f_('invalid_URL_{}').format(url))

        self.domain = match.group('domain').split('.')
        self.domain.reverse()
        self.path = [*filter(None, match.group('path').split('/'))]

class PatternMapping:
    """
    A mapping info, together with one of its patterns, as stored in Pattern
    Tree.
    """
    def __init__(self, pattern: str, mapping_info: MappingInfo):
        self.pattern = pattern
        self.mapping_info = mapping_info

    def register(self, pattern_tree: dict):
        """
        Make self queryable through the Pattern Tree passed in the argument.
        """
        deco = DeconstructedUrl(self.pattern)

        domain_tree = pattern_tree.get(deco.proto) or PatternTreeNode()
        pattern_tree[deco.proto] = domain_tree

        for path_tree in domain_tree.add(deco.domain, PatternTreeNode):
            for match_list in path_tree.add(deco.path, list):
                match_list.append(self)

class Malcontent:
    """
    Instance of this class represents a directory with files that can be loaded
    and served by Hydrilla.
    """
    def __init__(self, malcontent_dir_path: Path):
        """
        When an instance of Malcontent is constructed, it searches
        malcontent_dir_path for serveable site-modifying packages and loads
        them into its data structures.
        """
        self.infos = {'resource': {}, 'mapping': {}}
        self.pattern_tree = {}

        self.malcontent_dir_path = malcontent_dir_path

        if not self.malcontent_dir_path.is_dir():
            raise ValueError(f_('malcontent_dir_path_not_dir_{}')
                             .format(malcontent_dir_path))

        for item_type in ('mapping', 'resource'):
            type_path = self.malcontent_dir_path / item_type
            if not type_path.is_dir():
                continue

            for subpath in type_path.iterdir():
                if not subpath.is_dir():
                    continue

                for ver_file in subpath.iterdir():
                    try:
                        self._load_item(item_type, ver_file)
                    except Exception as e:
                        if flask.current_app._hydrilla_werror:
                            raise e from None

                        msg = f_('couldnt_load_item_from_{}').format(ver_file)
                        logging.error(msg, exc_info=True)

        self._report_missing()
        self._finalize()

    def _load_item(self, item_type: str, ver_file: Path) -> None:
        """
        Reads, validates and autocompletes serveable mapping/resource
        definition, then registers information from it in data structures.
        """
        version    = util.parse_version(ver_file.name)
        identifier = ver_file.parent.name

        item_json, major = util.load_instance_from_file(ver_file)

        util.validator_for(f'api_{item_type}_description-{major}.schema.json')\
            .validate(item_json)

        if item_type == 'resource':
            item_info = ResourceInfo(item_json, major)
        else:
            item_info = MappingInfo(item_json, major)

        if item_info.identifier != identifier:
            msg = f_('item_{item}_in_file_{file}')\
                .format({'item': item_info.identifier, 'file': ver_file})
            raise ValueError(msg)

        if item_info.version != version:
            ver_str = util.version_string(item_info.version)
            msg = f_('item_version_{ver}_in_file_{file}')\
                .format({'ver': ver_str, 'file': ver_file})
            raise ValueError(msg)

        versioned_info = self.infos[item_type].get(identifier)
        if versioned_info is None:
            versioned_info = VersionedItemInfo()
            self.infos[item_type][identifier] = versioned_info

        versioned_info.register(item_info)

    def _all_of_type(self, item_type: str) -> Iterable[ItemInfo]:
        """Iterator over all registered versions of all mappings/resources."""
        for versioned_info in self.infos[item_type].values():
            for item_info in versioned_info.by_version.values():
                yield item_info

    def _report_missing(self) -> None:
        """
        Use logger to print information about items that are referenced but
        were not loaded.
        """
        def report_missing_dependency(info: ResourceInfo, dep: str) -> None:
            msg = f_('no_dep_{resource}_{ver}_{dep}')\
                .format(dep=dep, resource=info.identifier,
                        ver=util.version_string(info.version))
            logging.error(msg)

        for resource_info in self._all_of_type('resource'):
            for dep in resource_info.dependencies:
                if dep not in self.infos['resource']:
                    report_missing_dependency(resource_info, dep)

        def report_missing_payload(info: MappingInfo, payload: str) -> None:
            msg = f_('no_payload_{mapping}_{ver}_{payload}')\
                .format(mapping=info.identifier, payload=payload,
                        ver=util.version_string(info.version))
            logging.error(msg)

        for mapping_info in self._all_of_type('mapping'):
            for payload in mapping_info.payloads.values():
                if payload not in self.infos['resource']:
                    report_missing_payload(mapping_info, payload)

        def report_missing_mapping(info: Union[MappingInfo, ResourceInfo],
                                   required_mapping: str) -> None:
            msg = _('no_mapping_{required_by}_{ver}_{required}')\
                .format(required_by=info.identifier, required=required_mapping,
                        ver=util.version_string(info.version))
            logging.error(msg)

        for item_info in (*self._all_of_type('mapping'),
                          *self._all_of_type('resource')):
            for required in item_info.required_mappings:
                if required not in self.infos['mapping']:
                    report_missing_mapping(item_info, required)

    def _finalize(self):
        """
        Initialize structures needed to serve queries. Called once after all
        data gets loaded.
        """
        for infos_dict in self.infos.values():
            for versioned_info in infos_dict.values():
                versioned_info.known_versions.sort()

        for info in self._all_of_type('mapping'):
            for pattern in info.payloads:
                try:
                    PatternMapping(pattern, info).register(self.pattern_tree)
                except Exception as e:
                    if flask.current_app._hydrilla_werror:
                        raise e from None
                    msg = f_('couldnt_register_{mapping}_{ver}_{pattern}')\
                        .format(mapping=info.identifier, pattern=pattern,
                                ver=util.version_string(info.version))
                    logging.error(msg)

    def query(self, url: str) -> list[MappingInfo]:
        """
        Return a list of registered mappings that match url.

        If multiple versions of a mapping are applicable, only the most recent
        is included in the result.
        """
        deco = DeconstructedUrl(url)

        collected = {}

        domain_tree = self.pattern_tree.get(deco.proto) or PatternTreeNode()

        def process_mapping(pattern_mapping: PatternMapping) -> None:
            if url[-1] != '/' and pattern_mapping.pattern[-1] == '/':
                return

            info = pattern_mapping.mapping_info

            if info.identifier not in collected or \
               info.version > collected[info.identifier].version:
                collected[info.identifier] = info

        for path_tree in domain_tree.search(deco.domain):
            for matches_list in path_tree.search(deco.path):
                for pattern_mapping in matches_list:
                    process_mapping(pattern_mapping)

        return list(collected.values())

bp = flask.Blueprint('bp', __package__)

class HydrillaApp(flask.Flask):
    """Flask app that implements a Hydrilla server."""
    def __init__(self, hydrilla_config: dict, flask_config: dict={}):
        """Create the Flask instance according to the configuration"""
        super().__init__(__package__, static_url_path='/',
                         static_folder=hydrilla_config['malcontent_dir'])
        self.config.update(flask_config)

        # https://stackoverflow.com/questions/9449101/how-to-stop-flask-from-initialising-twice-in-debug-mode
        if self.debug and os.environ.get('WERKZEUG_RUN_MAIN') != 'true':
            return

        self.jinja_options = {
            **self.jinja_options,
            'extensions': [
                *self.jinja_options.get('extensions', []),
                'jinja2.ext.i18n'
            ]
        }

        self._hydrilla_translation = \
            util.translation(here / 'locales', hydrilla_config['language'])
        self._hydrilla_project_url = hydrilla_config['hydrilla_project_url']
        self._hydrilla_port = hydrilla_config['port']
        self._hydrilla_werror = hydrilla_config.get('werror', False)

        if 'hydrilla_parent' in hydrilla_config:
            raise ValueError("Option 'hydrilla_parent' is not implemented.")

        malcontent_dir = Path(hydrilla_config['malcontent_dir']).resolve()
        with self.app_context():
            self._hydrilla_malcontent = Malcontent(malcontent_dir)

        self.register_blueprint(bp)

    def create_jinja_environment(self, *args, **kwargs) \
        -> flask.templating.Environment:
        """
        Flask's create_jinja_environment(), but tweaked to always include the
        'hydrilla_project_url' global variable and to install proper
        translations.
        """
        env = super().create_jinja_environment(*args, **kwargs)
        env.install_gettext_translations(self._hydrilla_translation)
        env.globals['hydrilla_project_url'] = self._hydrilla_project_url

        return env

    def run(self, *args, **kwargs):
        """
        Flask's run(), but tweaked to use the port from hydrilla configuration
        by default.
        """
        return super().run(*args, port=self._hydrilla_port, **kwargs)

def f_(text_key):
    return flask.current_app._hydrilla_translation.gettext(text_key)

def malcontent():
    return flask.current_app._hydrilla_malcontent

@bp.route('/')
def index():
    return flask.render_template('index.html')

identifier_json_re = re.compile(r'^([-0-9a-z.]+)\.json$')

def get_resource_or_mapping(item_type: str, identifier: str) -> Response:
    """
    Strip '.json' from 'identifier', look the item up and send its JSON
    description.
    """
    match = identifier_json_re.match(identifier)
    if not match:
        flask.abort(404)

    identifier = match.group(1)

    versioned_info = malcontent().infos[item_type].get(identifier)

    info = versioned_info and versioned_info.get_by_ver()
    if info is None:
        flask.abort(404)

    # no need for send_from_directory(); path is safe, constructed by us
    file_path = malcontent().malcontent_dir_path / item_type / info.path()
    return flask.send_file(open(file_path, 'rb'), mimetype='application/json')

@bp.route('/mapping/<string:identifier_dot_json>')
def get_newest_mapping(identifier_dot_json: str) -> Response:
    return get_resource_or_mapping('mapping', identifier_dot_json)

@bp.route('/resource/<string:identifier_dot_json>')
def get_newest_resource(identifier_dot_json: str) -> Response:
    return get_resource_or_mapping('resource', identifier_dot_json)

@bp.route('/query')
def query():
    url = flask.request.args['url']

    mapping_refs = [i.as_query_result() for i in malcontent().query(url)]
    result = {
        '$schema': 'https://hydrilla.koszko.org/schemas/api_query_result-1.schema.json',
        'mappings': mapping_refs,
        'generated_by': generated_by
    }

    return Response(json.dumps(result), mimetype='application/json')

@bp.route('/--help')
def mm_help():
    return start.get_help(click.Context(start_wsgi)) + '\n'

@bp.route('/--version')
def mm_version():
    prog_info = {'prog': 'Hydrilla', 'version': _version.version}
    return _('%(prog)s_%(version)s_license') % prog_info + '\n'

default_config_path = Path('/etc/hydrilla/config.json')
default_malcontent_dir = '/var/lib/hydrilla/malcontent'
default_project_url = 'https://hydrillabugs.koszko.org/projects/hydrilla/wiki'

console_gettext = util.translation(here / 'locales').gettext
_ = console_gettext

@click.command(help=_('serve_hydrilla_packages_explain_wsgi_considerations'))
@click.option('-m', '--malcontent-dir',
              type=click.Path(exists=True, file_okay=False),
              help=_('directory_to_serve_from_overrides_config'))
@click.option('-h', '--hydrilla-project-url', type=click.STRING,
              help=_('project_url_to_display_overrides_config'))
@click.option('-p', '--port', type=click.INT,
              help=_('tcp_port_to_listen_on_overrides_config'))
@click.option('-c', '--config', 'config_path',
              type=click.Path(exists=True, dir_okay=False, resolve_path=True),
              help=_('path_to_config_file_explain_default'))
@click.option('-l', '--language', type=click.STRING,
              help=_('language_to_use_overrides_config'))
@click.version_option(version=_version.version, prog_name='Hydrilla',
                      message=_('%(prog)s_%(version)s_license'),
                      help=_('version_printing'))
def start(malcontent_dir: Optional[str], hydrilla_project_url: Optional[str],
          port: Optional[int], config_path: Optional[str],
          language: Optional[str]) -> None:
    """
    Run a development Hydrilla server.

    This command is meant to be the entry point of hydrilla command exported by
    this package.
    """
    config_load_opts = {} if config_path is None \
        else {'config_path': [Path(config_path)]}

    hydrilla_config = config.load(**config_load_opts)

    if malcontent_dir is not None:
        hydrilla_config['malcontent_dir'] = str(Path(malcontent_dir).resolve())

    if hydrilla_project_url is not None:
        hydrilla_config['hydrilla_project_url'] = hydrilla_project_url

    if port is not None:
        hydrilla_config['port'] = port

    if language is not None:
        hydrilla_config['language'] = language

    lang = hydrilla_config.get('language')
    _ = console_gettext if lang is None else \
        util.translation(here / 'locales', lang).gettext

    for opt in ('malcontent_dir', 'hydrilla_project_url', 'port', 'language'):
        if opt not in hydrilla_config:
            raise ValueError(_('config_option_{}_not_supplied').format(opt))

    HydrillaApp(hydrilla_config).run()

@click.command(help=_('serve_hydrilla_packages_wsgi_help'),
               context_settings={
                   'ignore_unknown_options': True,
                   'allow_extra_args': True
               })
@click.version_option(version=_version.version, prog_name='Hydrilla',
                      message=_('%(prog)s_%(version)s_license'),
                      help=_('version_printing'))
def start_wsgi() -> None:
    """
    Create application object for use in WSGI deployment.

    This command Also handles --help and --version options in case it gets
    called outside WSGI environment.
    """
    return HydrillaApp(click.get_current_context().obj or config.load())



# SPDX-License-Identifier: 0BSD

# Copyright (C) 2013-2020, PyPA

# https://packaging.python.org/en/latest/guides/packaging-namespace-packages/#pkgutil-style-namespace-packages

__path__ = __import__('pkgutil').extend_path(__path__, __name__)



#!/usr/bin/env python3
# SPDX-License-Identifier: CC0-1.0

# Copyright (C) 2022 Wojtek Kosior <koszko@koszko.org>
#
# Available under the terms of Creative Commons Zero v1.0 Universal.

import setuptools

from setuptools.command.build_py import build_py
from setuptools.command.sdist import sdist

from pathlib import Path

here = Path(__file__).resolve().parent

class CustomBuildCommand(build_py):
    """The build command but runs babel before build."""
    def run(self, *args, **kwargs):
        """Wrapper around build_py's original run() method."""
        self.run_command('compile_catalog')

        super().run(*args, **kwargs)

class CustomSdistCommand(sdist):
    """
    The sdist command but prevents compiled message catalogs from being included
    in the archive.
    """
    def run(self, *args, **kwargs):
        """Wrapper around sdist's original run() method."""
        locales_dir = here / 'src/hydrilla/server/locales'
        locale_files = {}

        for path in locales_dir.rglob('*.mo'):
            locale_files[path] = path.read_bytes()

        for path in locale_files:
            path.unlink()

        super().run(*args, **kwargs)

        for path, contents in locale_files.items():
            path.write_bytes(contents)

setuptools.setup(cmdclass = {
    'build_py': CustomBuildCommand,
    'sdist': CustomSdistCommand
})



# SPDX-License-Identifier: CC0-1.0

# Copyright (C) 2022 Wojtek Kosior <koszko@koszko.org>
#
# Available under the terms of Creative Commons Zero v1.0 Universal.

import re

variable_word_re = re.compile(r'^<(.+)>$')

def process_command(command, expected_command):
    """Validate the command line and extract its variable parts (if any)."""
    assert len(command) == len(expected_command)

    extracted = {}
    for word, expected_word in zip(command, expected_command):
        match = variable_word_re.match(expected_word)
        if match:
            extracted[match.group(1)] = word
        else:
            assert word == expected_word

    return extracted

def run_missing_executable(command, **kwargs):
    """
    Instead of running a command, raise FileNotFoundError as if its executable
    was missing.
    """
    raise FileNotFoundError('dummy')

class MockedCompletedProcess:
    """
    Object with some fields similar to those of subprocess.CompletedProcess.
    """
    def __init__(self, args, returncode=0,
                 stdout='some output', stderr='some error output',
                 text_output=True):
        """
        Initialize MockedCompletedProcess. Convert strings to bytes if needed.
        """
        self.args       = args
        self.returncode = returncode

        if type(stdout) is str and not text_output:
            stdout = stdout.encode()
        if type(stderr) is str and not text_output:
            stderr = stderr.encode()

        self.stdout = stdout
        self.stderr = stderr
